CMD_GET_MEMORY = "procmem"
CMD_GET_MEMORY_PERC = "procmemp"
CMD_GET_CPU = "proccpu"
CMD_ADD_PROCESS = "addproc"
CMD_KILL_PROCESS = "killproc"
CMD_START_PROCESS = "startproc"
CMD_RESTART_PROCESS = "reproc"
CMD_REMOVE_PROCESS = "remproc"
CMD_LIST = "list"
CMD_STOP = "stop"
CMD_GET_PID = "procpid"
CMD_GET_UPTIME = "procupt"
CMD_GET_STDOUT = "procstdout"
CMD_GET_STDERR = "porcstderr"

DATA_CODE = b"\x00"
MSG_CODE = b"\x01"